#include "edimcoder.h"

void buffer_initEmptyBuffer(Buffer *buffer) {
    Line emptyLine;
    emptyLine.chars = NULL;
    
    Operation emptyOperation;
    emptyOperation.kind = Undo;
    emptyOperation.lines = NULL;
    emptyOperation.original = emptyLine;
    
    buffer->openedFilename = NULL;
    buffer->fileType = FT_UNKNOWN;
    buffer->lines = NULL;
    buffer->lastOperation = emptyOperation;
    buffer->modified = false;

#ifdef _WIN32
	buffer->lineEndingType = CR_LF;
#else
	buffer->lineEndingType = LF;
#endif

	buffer->indentSize = 0; // Use default
	buffer->useSpaces = defaultUseSpaces;

    buffer->outline.nodes = NULL;

    buffer->bookmarks = NULL;
}

// filename should be zero-terminated
// Returns 0 (false) if couldn't open file - however, the filetype is still set to the buffer.
int buffer_openFile(Buffer *buffer, char *filename) {
    FILE *fp;
	fp = fopen(filename, "rb");
    
    // Find last occurance of '.' in filename
    int dotIndex = -1;
    /*for (int i = strlen(filename) - 1; i >= 0; i++) {
        if (filename[i] == '.') {
            dotIndex = i;
            break;
        }
    }*/
    for (int i = 0; i < strlen(filename); i++) {
        if (filename[i] == '.') {
            dotIndex = i;
        }
    }
    
    // Determine filetype based on extension
    // TODO: Doesn't handle CPP files yet.
    FileType ft;
    if (dotIndex != -1) {
        switch (filename[dotIndex + 1]) {
            case 't':
            ft = FT_TEXT;
            break;
            case 'm':
            ft = FT_MARKDOWN;
            break;
            case 'c':
            ft = FT_C;
            break;
            case 'h':
            ft = FT_C_HEADER;
            break;
            case 'l':
            ft = FT_LUA;
            break;
            default:
            ft = FT_UNKNOWN;
            break;
        }
    } else {
        ft = FT_UNKNOWN;
    }
    if (ft != FT_UNKNOWN) {
        int isType = true;
        char *ftExt;
        getFileTypeExtension(ft, &ftExt);
        for (int i = dotIndex + 1; i < strlen(filename); i++) {
            if (filename[i] != ftExt[i - dotIndex - 1]) {
                isType = false;
                break;
            }
        }
        
        if (isType)
            buffer->fileType = ft;
        else buffer->fileType = FT_UNKNOWN;
        
        free(ftExt);
    }
    
    buffer->openedFilename = NULL;
    for (int i = 0; i < strlen(filename) + 1; i++) {
        buf_push(buffer->openedFilename, filename[i]);
    }
    
    // Make sure the filename ends with '\0'
    assert(buffer->openedFilename[buf_len(buffer->openedFilename) - 1] == '\0');
    char *chars = NULL;
    int line = 1;
    char c;
    
    // If fp is NULL, file doesn't exist. Return false after having set the fileType.
    if (fp == NULL) {
        buffer->modified = true;
        return false;
    }
    
	unsigned int lf_count = 0;
	unsigned int cr_lf_count = 0;
	unsigned int cr_count = 0;
	
	// Preliminary read to determine indent size
	int current_space_count = 0;
	int current_tab_count = 0;
	
	int *spaceNumbers = NULL; // Space Numbers List
	int *tabNumbers = NULL; // Tab Numbers List
	bool beginning = true;
	
	while ((c = fgetc(fp)) != EOF) {
		if (beginning && c == ' ') ++current_space_count;
		else if (beginning && c == '\t') ++current_tab_count;
		else if (c == '\n' || c == '\r') {
			buf_push(spaceNumbers, current_space_count);
			buf_push(tabNumbers, current_tab_count);
			current_space_count = 0;
			current_tab_count = 0;
			beginning = true;
		} else beginning = false;
	}
	
	// Determine which is used more, spaces vs. tabs
	int spacesUsed = 0;
	int tabsUsed = 0;
	
	for (int i = 0; i < buf_len(spaceNumbers); i++) {
		int spaceNum = spaceNumbers[i];
		int tabNum = tabNumbers[i];
		if (spaceNum > 0) ++spacesUsed;
		else if (tabNum > 0) ++tabsUsed;
	}
	
	if (spacesUsed > tabsUsed) {
		buffer->useSpaces = true;
	} else buffer->useSpaces = false;
	
	// Determine the smallest indent size (on a line that didn't have a tab)
	if (buffer->useSpaces) {
		int lowestSpaceNum = 0;
		if (buf_len(spaceNumbers) > 0) {
			int j = 0;
			while (lowestSpaceNum == 0) {
				lowestSpaceNum = spaceNumbers[j];
				++j;
			}
			
			for (int i = 0; i < buf_len(spaceNumbers); i++) {
				int spaceNum = spaceNumbers[i];
				int tabNum = tabNumbers[i];
				if (tabNum != 0) continue;
				if (spaceNum == 0) continue;
				
				if (spaceNum < lowestSpaceNum) lowestSpaceNum = spaceNum;
			}
		}
		
		buffer->indentSize = lowestSpaceNum == defaultIndentSize ? 0 : lowestSpaceNum;
	}
	
	buf_free(spaceNumbers);
	buf_free(tabNumbers);
	
	rewind(fp); // Reset to beginning of file
	
	// Read in characters to buffer, converting all types of line endings to '\n'
	// and counting the different types to see which is used most often (and will,
	// therefore, become the default during write/save)
	beginning = true;
    while ((c = fgetc(fp)) != EOF) {
        if (c == '\n') {
			++lf_count;
			buf_push(chars, '\n');
            buf_push(buffer->lines, ((Line) { chars }));
			beginning = true;
            ++line;
            chars = NULL; // Create new char Stretchy Buffer for next line.
        } else if (c == '\r') {
			c = fgetc(fp);
			if (c == EOF) {
				++cr_count;
				buf_push(chars, '\n');
				buf_push(buffer->lines, ((Line) { chars }));
				beginning = true;
				++line;
				chars = NULL;
				break;
			} else if (c == '\n') {
				++cr_lf_count;
				buf_push(chars, '\n');
				buf_push(buffer->lines, ((Line) { chars }));
				beginning = true;
				++line;
				chars = NULL;
			} else if (c == '\r') {
				cr_count += 2;
				buf_push(chars, '\n');
				buf_push(buffer->lines, ((Line) { chars }));
				++line;
				chars = NULL;
				buf_push(chars, '\n');
				buf_push(buffer->lines, ((Line) { chars }));
				beginning = true;
				++line;
				chars = NULL;
			} else {
				++cr_count;
				buf_push(chars, '\n');
				buf_push(buffer->lines, ((Line) { chars }));
				beginning = true;
				++line;
				chars = NULL;
				buf_push(chars, c);
			}
		} else if (beginning && c == ' ') {
			int count = 1;
			
			while ((c = fgetc(fp)) == ' ') ++count;
			fseek(fp, -1, SEEK_CUR); // Go back one character with the file pointer
			
			// Add appropriate amount of tabs based on indent size
			int tab_amt = count / (buffer->indentSize > 0 ? buffer->indentSize : defaultIndentSize);
			for (int i = 0; i < tab_amt; i++) {
				buf_push(chars, '\t');
			}
			
			// Add rest of spaces
			int space_amt = count % (buffer->indentSize > 0 ? buffer->indentSize : defaultIndentSize);
			for (int i = 0; i < space_amt; i++) {
				buf_push(chars, ' ');
			}
/*		} else if (c == ' ' || c == '\t') {
			if (c == ' ') ++indent_space_count;
			if (c == '\t') ++indent_tab_count;
			
			int current_tab_count = (c == '\t' ? 1 : 0);
			int current_space_count = c == ' ';
			
			while ((c = fgetc()) == ' ' || c == '\t') {
				if (c == ' ') {
					++current_space_count;
				} else if (c == '\t') {
					++current_tab_count;
				}
			}
		*/
		} else {
			buf_push(chars, c);
			if (c != '\t')
				beginning = false;
		}
    }
    
    fclose(fp);

	// See which line ending type was used most often, and set to default for write/save
	if (lf_count > cr_lf_count) {
		if (lf_count > cr_count) {
			buffer->lineEndingType = LF;
		} else {
			buffer->lineEndingType = CR;
		}
	} else {
		if (cr_lf_count > cr_count) {
			buffer->lineEndingType = CR_LF;
		} else {
			buffer->lineEndingType = CR;
		}
	}
    
    // Set modified to false and current line to last line in file.
    buffer->modified = false;
    buffer->currentLine = buf_len(buffer->lines);
    
    // Create the outline
    createOutline();
    
    return true;
}

void buffer_close(Buffer *buffer) {
    // Clear openedFilename and the file information
    buf_free(buffer->openedFilename);
    
    for (int i = 0; i < buf_len(buffer->lines); i++) {
        buf_free(buffer->lines[i].chars);
    }

    // Clear the bookmarks (and names)
    for (int i = 0; i < buf_len(buffer->bookmarks); i++) {
        buf_free(buffer->bookmarks[i].name);
    }
    buf_free(buffer->bookmarks);
    
    // Clear the outline
    switch (buffer->fileType) {
        case FT_MARKDOWN:
        {
            buf_free(buffer->outline.markdown_nodes);
        } break;
        case FT_C:
        {
            buf_free(buffer->outline.c_nodes);
        } break;;
    }
    
    // Free the buffer
    buf_free(buffer->lines);
}

// If openedFilename is not set in the buffer, then filename is used.
// TODO: Boolean to forcibly use the filename passed in (useful for saveas)
void buffer_saveFile(Buffer *buffer, char *filename) {
    FILE *fp;
    if (buffer->openedFilename && buf_len(buffer->openedFilename) > 0) {
        fp = fopen(buffer->openedFilename, "wb");
    } else {
        fp = fopen(filename, "wb");
        // Copy filename into openedFilename
        for (int i = 0; i < strlen(filename) + 1; i++) {
            buf_push(buffer->openedFilename, filename[i]);
        }
        
        // Find last occurance of '.' in filename
        int dotIndex = -1;
        for (int i = 0; i < strlen(filename); i++) {
            if (filename[i] == '.') {
                dotIndex = i;
            }
        }
        
        // Determine filetype based on extension
        // TODO: Doesn't handle CPP files yet.
        FileType ft;
        if (dotIndex != -1) {
            switch (filename[dotIndex + 1])
            {
                case 't':
                ft = FT_TEXT;
                break;
                case 'm':
                ft = FT_MARKDOWN;
                break;
                case 'c':
                ft = FT_C;
                break;
                case 'h':
                ft = FT_C_HEADER;
                break;
                case 'l':
                ft = FT_LUA;
                break;
                default:
                ft = FT_UNKNOWN;
                break;
            }
        } else {
            ft = FT_UNKNOWN;
        }
        if (ft != FT_UNKNOWN) {
            int isType = true;
            char *ftExt;
            getFileTypeExtension(ft, &ftExt);
            for (int i = dotIndex + 1; i < strlen(filename); i++) {
                if (filename[i] != ftExt[i - dotIndex - 1]) {
                    isType = false;
                    break;
                }
            }
            
            if (isType) {
                buffer->fileType = ft;
            } else {
                buffer->fileType = FT_UNKNOWN;
            }
            free(ftExt);
        }
    }
    
    // Write the characters out to the file
    for (int line = 0; line < buf_len(buffer->lines); line++) {
        for (int i = 0; i < buf_len(buffer->lines[line].chars); i++) {
			if (buffer->lines[line].chars[i] == '\n') {
				// Convert \n to correct line ending based on line ending type of file
				if (buffer->lineEndingType == CR_LF) {
					fputc('\r', fp);
					fputc('\n', fp);
				} else if (buffer->lineEndingType == LF) {
					fputc('\n', fp);
				} else if (buffer->lineEndingType == CR) {
					fputc('\r', fp);
				}
			} else if (buffer->lines[line].chars[i] == '\t') {
				if (buffer->useSpaces) {
					for (int i = 0; i < (buffer->indentSize > 0 ? buffer->indentSize : defaultIndentSize); i++)
						fputc(' ', fp);
				} else {
					fputc('\t', fp);
				}
			} else {
				fputc(buffer->lines[line].chars[i], fp);
			}
        }
    }
    
    buffer->modified = false;
    
    fclose(fp);
}

// The lines buffer isn't freed.
// Returns the line after the lines that were added.
int buffer_insertAfterLine(Buffer *buffer, int line, Line *lines) {
    int lineToInsertAfter = line;
    if (line == -1) {
        lineToInsertAfter = buffer->currentLine;
    }
    
    int linesAddedAmt = buf_len(lines);
    int amtToMove = buf_len(buffer->lines) - line;
    
    // Add space for the new lines into the buffer
    buf_add(buffer->lines, linesAddedAmt);
    
    // Move the lines after the line inserting after up by how many lines were inserted // TODO: Make this message clearer
    Line *moveSource = &(buffer->lines[lineToInsertAfter - 1 + 1]);
    Line *moveDestination = moveSource + linesAddedAmt;
    memmove(moveDestination, moveSource, amtToMove * sizeof(Line));
    
    // Copy the lines that were inserted to directly after the line inserting after
    Line *copySource = lines;
    Line *copyDestination = moveSource;
    size_t copyAmt = linesAddedAmt * sizeof(Line);
    memcpy(copyDestination, copySource, copyAmt);
    
    // Set cursor to the last line that was inserted
    buffer->modified = true;
    buffer->currentLine = lineToInsertAfter + linesAddedAmt;
    
    // Return the first line that was moved to make room for the inserted lines
    return buffer->currentLine + 1;
}

// The lines buffer isn't freed.
// Returns the line after the lines that were added.
int buffer_insertBeforeLine(Buffer *buffer, int line, Line *lines) {
    int lineToInsertBefore = line;
    if (line == -1 || line == 0) {
        lineToInsertBefore = buffer->currentLine;
        if (lineToInsertBefore == 0)
            lineToInsertBefore = 1;
    }
    
    int linesAddedAmt = buf_len(lines);
    int amtToMove = buf_len(buffer->lines) - (line - 1);
    
    // Add space for the new lines into the buffer
    buf_add(buffer->lines, linesAddedAmt);
    
    // Move the lines after the passed in line (and including it) up by how many lines are being inserted before
    Line *moveSource = &(buffer->lines[lineToInsertBefore - 1]);
    Line *moveDestination = moveSource + linesAddedAmt;
    memmove(moveDestination, moveSource, amtToMove * sizeof(Line));
    
    // Copy the lines that were inserted to the newly created space
    Line *copySource = lines;
    Line *copyDestination = moveSource;
    size_t copyAmt = linesAddedAmt * sizeof(Line);
    memcpy(copyDestination, copySource, copyAmt);
    
    // Set the current line to the line that the lines were inserted before
    buffer->modified = true;
    buffer->currentLine = lineToInsertBefore + linesAddedAmt;
    
    // Return the first line that was moved to make room for the inserted lines
    return buffer->currentLine;
}

// Pass in a char buffer to append to line
// Does not free the chars buffer
void buffer_appendToLine(Buffer *buffer, int line, char *chars) {
    int lineToAppendTo = line;
    if (line == -1 || line == 0) {
        lineToAppendTo = buffer->currentLine;
        if (lineToAppendTo == 0)
            return;
    }
    // Remove the new line character from the line
    buf_pop(buffer->lines[lineToAppendTo - 1].chars);
    
    // Copy from the passed-in chars buffer to the line's chars buffer
    size_t num = buf_len(chars);
    char *destination = buf_add(buffer->lines[lineToAppendTo - 1].chars, num);
    strncpy(destination, chars, num);
    
    buffer->modified = true;
    buffer->currentLine = lineToAppendTo;
}

// Pass in a char buffer that will be used in place of the new line. This buffer should not end in a new line.
// The old char buffer of the line will be freed.
void buffer_prependToLine(Buffer *buffer, int line, char *chars) {
    int lineToPrependTo = line;
    if (line == -1 || line == 0) {
        lineToPrependTo = buffer->currentLine;
        if (lineToPrependTo == 0)
            return;
    }
    
    // Store the old buffer
    char *oldBuffer = buffer->lines[lineToPrependTo - 1].chars;
    
    // Set the line to the new buffer passed in
    buffer->lines[lineToPrependTo - 1].chars = chars;
    
    // Push onto the buffer the chars of the old buffer
    size_t num = buf_len(oldBuffer);
    char *destination = buf_add(buffer->lines[lineToPrependTo - 1].chars, num);
    strncpy(destination, oldBuffer, num);
    
    // Free the old buffer
    buf_free(oldBuffer);
    
    buffer->modified = true;
    buffer->currentLine = lineToPrependTo;
}

// Pass in a char buffer that the line's char buffer will be replaced with. This buffer should likely end in a new line.
// The old char buffer of the line will be freed.
void buffer_replaceLine(Buffer *buffer, int line, char *chars) {
    int lineToReplace = line;
    if (line == -1 || line == 0) {
        lineToReplace = buffer->currentLine;
        if (lineToReplace == 0)
            return;
    }
    
    // Free the old buffer
    buf_free(buffer->lines[lineToReplace - 1].chars);
    
    // Set the line to the new buffer passed in
    buffer->lines[lineToReplace - 1].chars = chars;
    
    buffer->modified = true;
    buffer->currentLine = lineToReplace;
}

// Replace in the line from the startIndex to the endIndex (inclusive) with the provided char buffer
void buffer_replaceInLine(Buffer *buffer, int line, int startIndex, int endIndex, char *chars) {
    int lineToReplaceIn = line;
    if (line == -1 || line == 0) {
        lineToReplaceIn = buffer->currentLine;
        if (lineToReplaceIn == 0)
            return;
    }
    
	int lengthOfStringToReplace = endIndex - startIndex + 1;
	int addedAmt = buf_len(chars) - lengthOfStringToReplace;
    int amtToMove = buf_len(buffer->lines[lineToReplaceIn - 1].chars) - endIndex;
    
    // If replacement string is bigger than string to replace, add characters to the buffer. Otherwise, if replacement string is smaller, pop off the correct amount
    if (addedAmt > 0)
        buf_add(buffer->lines[lineToReplaceIn - 1].chars, addedAmt);
    else if (addedAmt < 0) {
        for (int i = addedAmt; i < 0; i++) {
            buf_pop(buffer->lines[lineToReplaceIn - 1].chars);
        }
    }
    
    // Move over the characters due to the replacement string being bigger/smaller than the string being replaced
    char *moveSource = &(buffer->lines[lineToReplaceIn - 1].chars[endIndex + 1]);
    char *moveDestination = &(buffer->lines[lineToReplaceIn - 1].chars[endIndex + 1 + addedAmt]);
    size_t bytes = sizeof(char) * amtToMove;
    memmove(moveDestination, moveSource, bytes);
    
    // Copy over the replacement string
    char *source = chars;
    char *destination = &(buffer->lines[lineToReplaceIn - 1].chars[startIndex]);
    size_t copyAmt = buf_len(chars) * sizeof(char);
    memcpy(destination, source, copyAmt);
    
    buffer->modified = true;
    buffer->currentLine = lineToReplaceIn;
}

// TODO: Add the number to move up by
void buffer_moveLineUp(Buffer *buffer, int line) {
    int lineToMove = line;
    if (line == -1 || line == 0) {
        lineToMove = buffer->currentLine;
        if (lineToMove == 0)
            return;
    }
    
    // Store the line where the given line is being moved up to
    Line tmp = buffer->lines[lineToMove - 2];
    
    // Set the new position of the line being moved
    buffer->lines[lineToMove - 2] = buffer->lines[lineToMove - 1];
    
    // Set the old position to the line stored in tmp (the line that's moved down)
    buffer->lines[lineToMove - 1] = tmp;
    
    // Set the currentLine to the new position of the line that was moved up
    buffer->modified = true;
    buffer->currentLine = lineToMove - 1;
}

// TODO: Add the number to move up by
void buffer_moveLineDown(Buffer *buffer, int line) {
    int lineToMove = line;
    if (line == -1 || line == 0) {
        lineToMove = buffer->currentLine;
        if (lineToMove == 0)
            return;
    }
    
    // Store the line where the given line is being moved down to
    Line tmp = buffer->lines[lineToMove];
    
    // Set the new position of the line being moved
    buffer->lines[lineToMove] = buffer->lines[lineToMove - 1];
    
    // Set the old position to the line stored in tmp (the line that's moved up)
    buffer->lines[lineToMove - 1] = tmp;
    
    // Set the currentLine to the new position of the line that was moved down
    buffer->modified = true;
    buffer->currentLine = lineToMove + 1;
}

void buffer_deleteLine(Buffer *buffer, int line) {
    int lineToDelete = line;
    if (line == -1 || line == 0) {
        lineToDelete = buffer->currentLine;
        if (lineToDelete == 0)
            return;
    }
    
    if (lineToDelete == buf_len(buffer->lines)) {
        buf_free(buffer->lines[buf_len(buffer->lines) - 1].chars);
        buf_pop(buffer->lines);
    } else {
        // Delete the char buffer of the line that's being deleted
        buf_free(buffer->lines[lineToDelete - 1].chars);
        
        // Move all lines down one
        void *source = &(buffer->lines[lineToDelete - 1 + 1]);
        void *destination = &(buffer->lines[lineToDelete - 1]);
        size_t bytes = sizeof(Line) * (buf_len(buffer->lines) - lineToDelete);
        memmove(destination, source, bytes);
        
        // Decrease the length of the buffer, keeping the char buffer of this last line because is was moved down one.
        buf_pop(buffer->lines);
    }
    
    // Set the cursor the the line that was deleted
    buffer->modified = true;
    if (lineToDelete > buf_len(buffer->lines))
        buffer->currentLine = buf_len(buffer->lines);
    else buffer->currentLine = lineToDelete;
}

// The string may be or not be a stretchy buffer. The length should be passed in.
// Returns index of first occurance of string, -1 for no occurance
int buffer_findStringInLine(Buffer *buffer, int line, char *str, int strLength, bool insensitive) {
    int lineToSearch = line;
    if (line == -1 || line == 0) {
        lineToSearch = buffer->currentLine;
        if (lineToSearch == 0)
            return -1;
    }
    // Find the first occurance of the string in the current line
    int index = -1; // Column index
    int ii = 0;
    
    for (int i = 0; i < buf_len(buffer->lines[lineToSearch - 1].chars); i++) {
		bool str_isUppercase = str[ii] >= 'A' && str[ii] <= 'Z';
		bool str_isLowercase = str[ii] >= 'a' && str[ii] <= 'z';
			
		if ((buffer->lines[lineToSearch - 1].chars[i] == str[ii])
			|| (insensitive && str_isUppercase && buffer->lines[lineToSearch - 1].chars[i] == str[ii] + 32)
			|| (insensitive && str_isLowercase && buffer->lines[lineToSearch - 1].chars[i] == str[ii] - 32)) {
            if (ii == 0)
                index = i;
            ++ii;
        } else {
            ii = 0;
            index = -1;
        }
        
        // If string ends in a new line or 0 termination, subtract one from the string length so we don't match them
        if (str[strLength - 1] == '\0' || str[strLength - 1] == '\n') {
            if (ii == strLength - 1) break;
        } else {
            if (ii == strLength) break;
        }
    }
    
    buffer->currentLine = lineToSearch;
    
    // Return the index
    return index;
}

// The string must not be null terminated and can be or not be a stretchy buffer. The length should be passed in.
// Returns the index to the line where the string was found. Also sets the column index, that was passed in, to the index of the first occurance in that line (this index counts from 0).
int buffer_findString(Buffer *buffer, char *str, int strLength, int *colIndex, unsigned int startLine, unsigned int endLine, bool insensitive) {
    // Go though all lines in buffer
	for (int line = startLine; line < endLine; line++) {
        int index = buffer_findStringInLine(buffer, line + 1, str, strLength, insensitive);
        
        if (index != -1) {
            (*colIndex) = index;
            buffer->currentLine = line + 1; // TODO
            return line;
        }
    }
    
    return -1;
}
